from __future__ import annotations

import json
import os
import random
import string
import time
from datetime import datetime
from urllib.parse import quote

from ..types import (
    ProviderName,
    SandboxStatus,
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)
from ..config import ProviderConfig
from ..provider import Provider, Sandbox
from ..errors import NotFoundError, NotSupportedError, ProviderError, BadConfigError
from .._http_client import HttpClient, HTTPError
from .._util import first_non_empty
from ..registry import register_provider

_DEFAULT_CONTROL = "https://api.blaxel.ai/v0"


def _elapsed_ms(start: float) -> int:
    return int((time.monotonic() - start) * 1000)


def _encode_fs_path(path: str) -> str:
    norm = path.lstrip("/")
    if not norm:
        return ""
    return "/".join(quote(seg, safe="") for seg in norm.split("/"))


def _normalize_base_url(url: str | None) -> str:
    if not url:
        return ""
    return url.rstrip("/")


def _strip_name_labels(meta: dict[str, str]) -> dict[str, str]:
    return {k: v for k, v in meta.items() if k not in ("name", "sandboxName", "createIfNotExist")}


def _map_deployment_status(s: str | None) -> SandboxStatus:
    u = (s or "").upper()
    if u == "DEPLOYED":
        return SandboxStatus.RUNNING
    if u in ("DEPLOYING", "BUILDING", "UPLOADING"):
        return SandboxStatus.STARTING
    if u in ("DEACTIVATED", "TERMINATED", "DELETING", "DEACTIVATING"):
        return SandboxStatus.STOPPED
    if u == "FAILED":
        return SandboxStatus.ERROR
    return SandboxStatus.RUNNING


def _parse_started_at(raw: str | None) -> datetime:
    if not raw:
        return datetime.now()
    try:
        if raw.endswith("Z"):
            raw = raw[:-1] + "+00:00"
        return datetime.fromisoformat(raw)
    except ValueError:
        return datetime.now()


def _sandbox_row_to_info(row: dict, sid: str) -> SandboxInfo:
    meta = row.get("metadata") or {}
    spec = row.get("spec") or {}
    runtime = spec.get("runtime") or {}
    mem = runtime.get("memory")
    info = SandboxInfo(
        id=sid,
        provider=ProviderName.BLAXEL,
        status=_map_deployment_status(row.get("status")),
        started_at=_parse_started_at(meta.get("createdAt")),
        metadata=meta.get("labels") or {},
        template=runtime.get("image"),
    )
    if isinstance(mem, int) and mem > 0:
        info.memory_mb = mem
        info.cpus = round(mem / 2048)
    return info


def _map_err(err: HTTPError) -> Exception:
    if err.status == 404:
        return NotFoundError(str(err))
    msg = err.body.decode(errors="replace")
    try:
        j = json.loads(msg)
        if isinstance(j, dict):
            if j.get("message"):
                msg = str(j["message"])
            elif j.get("error"):
                msg = str(j["error"])
    except Exception:
        pass
    return ProviderError(provider="blaxel", message=msg, status_code=err.status)


def _latin1_from_bytes(buf: bytes) -> str:
    return buf.decode("latin-1")


class BlaxelProvider(Provider):
    def __init__(self, cfg: ProviderConfig) -> None:
        tok = first_non_empty(
            cfg.api_key,
            os.environ.get("BLAXEL_API_KEY"),
            os.environ.get("BL_API_KEY"),
            os.environ.get("SANDBOXER_API_KEY"),
        )
        if not tok:
            raise BadConfigError("Blaxel API key required (api_key, BLAXEL_API_KEY, BL_API_KEY, or SANDBOXER_API_KEY)")
        timeout_s = 30.0
        if cfg.default_timeout_ms:
            timeout_s = cfg.default_timeout_ms / 1000.0
        self._hc = HttpClient(timeout_s=timeout_s)
        self._token = tok
        self._control_base = (cfg.base_url or os.environ.get("BLAXEL_API_BASE") or _DEFAULT_CONTROL).rstrip("/")
        ws = cfg.extra.get("workspace") if isinstance(cfg.extra.get("workspace"), str) else None
        self._workspace = first_non_empty(
            ws or "",
            os.environ.get("BLAXEL_WORKSPACE") or "",
            os.environ.get("BL_WORKSPACE") or "",
        )

    def _control_headers(self) -> dict[str, str]:
        h = {"Authorization": f"Bearer {self._token}"}
        if self._workspace:
            h["X-Blaxel-Workspace"] = self._workspace
        return h

    def _sandbox_headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self._token}"}

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        filter = filter or ListSandboxesFilter()
        if filter.provider is not None and filter.provider != ProviderName.BLAXEL:
            return []
        url = f"{self._control_base}/sandboxes"
        try:
            rows = self._hc.do("GET", url, self._control_headers())
        except HTTPError as e:
            raise _map_err(e) from e
        if not rows or not isinstance(rows, list):
            return []
        out: list[SandboxInfo] = []
        for s in rows:
            if not isinstance(s, dict):
                continue
            meta = s.get("metadata") or {}
            name = meta.get("name")
            if not name:
                continue
            if filter.metadata_filter:
                labels = meta.get("labels") or {}
                hay = " ".join([*labels.values(), name])
                if filter.metadata_filter not in hay:
                    continue
            out.append(_sandbox_row_to_info(s, name))
            if filter.limit and filter.limit > 0 and len(out) >= filter.limit:
                break
        return out

    def kill_sandbox(self, sandbox_id: str) -> None:
        url = f"{self._control_base}/sandboxes/{quote(sandbox_id, safe='')}"
        try:
            self._hc.do("DELETE", url, self._control_headers())
        except HTTPError as e:
            raise _map_err(e) from e

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        req = req or CreateSandboxRequest()
        meta = req.metadata or {}
        name = meta.get("name") or meta.get("sandboxName")
        if not name:
            suffix = "".join(random.choices(string.ascii_lowercase + string.digits, k=6))
            name = f"sandboxer-{int(time.time() * 1000):x}-{suffix}"

        runtime: dict = {"image": req.template or "blaxel/base-image:latest"}
        if req.memory_mb is not None:
            runtime["memory"] = req.memory_mb
        elif req.cpus is not None:
            runtime["memory"] = req.cpus * 2048
        else:
            runtime["memory"] = 4096

        if req.envs:
            runtime["envs"] = [{"name": k, "value": v} for k, v in req.envs.items()]

        md: dict = {"name": name}
        labs = _strip_name_labels(dict(meta)) if meta else {}
        if labs:
            md["labels"] = labs
        body: dict = {"metadata": md, "spec": {"runtime": runtime}}
        q = "?createIfNotExist=true" if meta.get("createIfNotExist") == "true" else ""
        url = f"{self._control_base}/sandboxes{q}"
        try:
            created = self._hc.do("POST", url, self._control_headers(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        if not isinstance(created, dict):
            raise ProviderError("blaxel", "invalid create sandbox response")

        base_url = _normalize_base_url((created.get("metadata") or {}).get("url"))
        if not base_url:
            get_url = f"{self._control_base}/sandboxes/{quote(name, safe='')}"
            try:
                again = self._hc.do("GET", get_url, self._control_headers())
            except HTTPError as e:
                raise _map_err(e) from e
            if isinstance(again, dict):
                base_url = _normalize_base_url((again.get("metadata") or {}).get("url"))
        if not base_url:
            raise ProviderError("blaxel", "sandbox created but no endpoint URL in response (metadata.url)")

        sb = BlaxelSandbox(self, name, base_url)
        info = _sandbox_row_to_info(created, name)
        return sb, info

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        url = f"{self._control_base}/sandboxes/{quote(sandbox_id, safe='')}"
        try:
            row = self._hc.do("GET", url, self._control_headers())
        except HTTPError as e:
            raise _map_err(e) from e
        if not isinstance(row, dict):
            raise ProviderError("blaxel", "invalid get sandbox response")
        meta = row.get("metadata") or {}
        name = meta.get("name") or sandbox_id
        base_url = _normalize_base_url(meta.get("url"))
        if not base_url:
            raise ProviderError("blaxel", "sandbox has no endpoint URL (metadata.url); cannot attach")
        return BlaxelSandbox(self, name, base_url)

    def close(self) -> None:
        self._hc.close()


class BlaxelSandbox(Sandbox):
    def __init__(self, provider: BlaxelProvider, sandbox_id: str, base_url: str) -> None:
        self._p = provider
        self._id = sandbox_id
        self._base = base_url.rstrip("/")

    @property
    def id(self) -> str:
        return self._id

    def _fs_url(self, path: str) -> str:
        enc = _encode_fs_path(path)
        return f"{self._base}/filesystem/{enc}"

    def info(self) -> SandboxInfo:
        url = f"{self._p._control_base}/sandboxes/{quote(self._id, safe='')}"
        try:
            row = self._p._hc.do("GET", url, self._p._control_headers())
        except HTTPError as e:
            raise _map_err(e) from e
        if not isinstance(row, dict):
            raise ProviderError("blaxel", "invalid sandbox info response")
        return _sandbox_row_to_info(row, self._id)

    def is_running(self) -> bool:
        i = self.info()
        return i.status in (SandboxStatus.RUNNING, SandboxStatus.STARTING)

    def pause(self) -> None:
        raise NotSupportedError("blaxel: pause not supported by API")

    def resume(self) -> None:
        raise NotSupportedError("blaxel: resume not supported by API")

    def kill(self) -> None:
        self._p.kill_sandbox(self._id)

    def port_url(self, port: int) -> str:
        return f"{self._base}/port/{port}"

    def run_command(self, req: RunCommandRequest) -> CommandResult:
        start = time.monotonic()
        body: dict = {"command": req.cmd, "waitForCompletion": True}
        if req.cwd:
            body["workingDir"] = req.cwd
        if req.timeout_seconds is not None:
            body["timeout"] = req.timeout_seconds
        if req.env:
            body["env"] = req.env
        url = f"{self._base}/process"
        try:
            out = self._p._hc.do("POST", url, self._p._sandbox_headers(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        if not isinstance(out, dict):
            out = {}
        code = int(out.get("exitCode") or 0)
        return CommandResult(
            stdout=str(out.get("stdout") or ""),
            stderr=str(out.get("stderr") or ""),
            exit_code=code,
            duration_ms=_elapsed_ms(start),
        )

    def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        body: dict = {"command": req.cmd, "waitForCompletion": False}
        if req.cwd:
            body["workingDir"] = req.cwd
        if req.env:
            body["env"] = req.env
        url = f"{self._base}/process"
        try:
            out = self._p._hc.do("POST", url, self._p._sandbox_headers(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        if not isinstance(out, dict):
            raise ProviderError("blaxel", "invalid start command response")
        handle = str(out.get("pid") or "")
        if not handle or not handle.isdigit():
            raise ProviderError("blaxel", "process start did not return pid")
        return int(handle, 10), handle

    def wait_for_handle(self, handle_id: str) -> CommandResult:
        start = time.monotonic()
        deadline = time.monotonic() + 3600.0
        while time.monotonic() < deadline:
            url = f"{self._base}/process/{quote(handle_id, safe='')}"
            try:
                last = self._p._hc.do("GET", url, self._p._sandbox_headers())
            except HTTPError as e:
                raise _map_err(e) from e
            if not isinstance(last, dict):
                time.sleep(0.4)
                continue
            st = str(last.get("status") or "").lower()
            if st in ("completed", "failed", "killed", "stopped"):
                code = int(last.get("exitCode") or (0 if st == "completed" else 1))
                return CommandResult(
                    stdout=str(last.get("stdout") or ""),
                    stderr=str(last.get("stderr") or ""),
                    exit_code=code,
                    duration_ms=_elapsed_ms(start),
                )
            time.sleep(0.4)
        raise ProviderError("blaxel", "wait_for_handle: timeout waiting for process")

    def kill_process(self, pid: int) -> None:
        url = f"{self._base}/process/{quote(str(pid), safe='')}/kill"
        try:
            self._p._hc.do("DELETE", url, self._p._sandbox_headers())
        except HTTPError as e:
            raise _map_err(e) from e

    def list_processes(self) -> list[ProcessInfo]:
        url = f"{self._base}/process"
        try:
            rows = self._p._hc.do("GET", url, self._p._sandbox_headers())
        except HTTPError as e:
            raise _map_err(e) from e
        if not rows or not isinstance(rows, list):
            return []
        out: list[ProcessInfo] = []
        for p in rows:
            if not isinstance(p, dict):
                continue
            pid_s = str(p.get("pid") or "0")
            try:
                pid = int(pid_s, 10)
            except ValueError:
                pid = 0
            out.append(ProcessInfo(pid=pid, command=str(p.get("command") or "")))
        return out

    def read_file(self, path: str) -> bytes:
        url = self._fs_url(path)
        h = dict(self._p._sandbox_headers())
        h["Accept"] = "application/octet-stream,*/*"
        try:
            _, raw = self._p._hc.do_raw("GET", url, h)
        except HTTPError as e:
            if e.status == 404:
                raise NotFoundError(str(e)) from e
            raise _map_err(e) from e
        return raw

    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        _ = user
        body: dict = {"content": _latin1_from_bytes(content)}
        if mode is not None:
            body["permissions"] = oct(mode & 0o777)[2:].zfill(3)
        try:
            self._p._hc.do("PUT", self._fs_url(path), self._p._sandbox_headers(), body)
        except HTTPError as e:
            raise _map_err(e) from e

    def list_directory(self, path: str) -> list[FileInfo]:
        try:
            d = self._p._hc.do("GET", self._fs_url(path), self._p._sandbox_headers())
        except HTTPError as e:
            raise _map_err(e) from e
        if not isinstance(d, dict):
            return []
        out: list[FileInfo] = []
        for f in d.get("files") or []:
            if not isinstance(f, dict):
                continue
            out.append(
                FileInfo(
                    name=str(f.get("name") or ""),
                    path=str(f.get("path") or ""),
                    is_dir=False,
                    size=int(f.get("size") or 0),
                )
            )
        for sub in d.get("subdirectories") or []:
            if not isinstance(sub, dict):
                continue
            out.append(
                FileInfo(
                    name=str(sub.get("name") or ""),
                    path=str(sub.get("path") or ""),
                    is_dir=True,
                    size=0,
                )
            )
        return out

    def make_dir(self, path: str) -> None:
        try:
            self._p._hc.do("PUT", self._fs_url(path), self._p._sandbox_headers(), {"isDirectory": True})
        except HTTPError as e:
            raise _map_err(e) from e

    def remove(self, path: str) -> None:
        url = f"{self._fs_url(path)}?recursive=true"
        try:
            self._p._hc.do("DELETE", url, self._p._sandbox_headers())
        except HTTPError as e:
            raise _map_err(e) from e

    def exists(self, path: str) -> bool:
        try:
            self._p._hc.do_raw("GET", self._fs_url(path), self._p._sandbox_headers())
        except HTTPError as e:
            if e.status == 404:
                return False
            raise _map_err(e) from e
        return True

    def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        raise NotSupportedError("blaxel: PTY not exposed in sandbox API")

    def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        raise NotSupportedError("blaxel: PTY not exposed in sandbox API")

    def kill_pty(self, pid: int) -> None:
        raise NotSupportedError("blaxel: PTY not exposed in sandbox API")

    def list_pty(self) -> list[PTYInfo]:
        raise NotSupportedError("blaxel: PTY not exposed in sandbox API")


def _make_blaxel_provider(cfg: ProviderConfig) -> Provider:
    return BlaxelProvider(cfg)


register_provider(ProviderName.BLAXEL, _make_blaxel_provider)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass
class ProviderConfig:
    api_key: str | None = None
    base_url: str | None = None
    default_timeout_ms: int | None = None
    extra: dict[str, Any] = field(default_factory=dict)



from __future__ import annotations


def normalize_path(p: str) -> str:
    return p.strip()


def first_non_empty(*vals: str | None) -> str:
    for v in vals:
        if v:
            return v
    return ""


def shell_quote(p: str) -> str:
    return "'" + p.replace("'", "'\"'\"'") + "'"



from __future__ import annotations

import httpx


class HTTPError(Exception):
    def __init__(self, status: int, body: bytes):
        super().__init__(f"http status {status}")
        self.status = status
        self.body = body


class HttpClient:
    def __init__(self, timeout_s: float = 30.0):
        self._client = httpx.Client(timeout=timeout_s)

    def do(self, method: str, url: str, headers: dict[str, str], body: dict | None = None) -> dict | list | None:
        kwargs: dict = {"method": method, "url": url, "headers": headers}
        if body is not None:
            kwargs["json"] = body
        resp = self._client.request(**kwargs)
        if resp.status_code >= 400:
            raise HTTPError(resp.status_code, resp.content)
        if not resp.content:
            return None
        return resp.json()

    def do_raw(
        self,
        method: str,
        url: str,
        headers: dict[str, str],
        content: bytes | None = None,
        content_type: str | None = None,
    ) -> tuple[int, bytes]:
        h = dict(headers)
        if content_type:
            h["Content-Type"] = content_type
        resp = self._client.request(method=method, url=url, headers=h, content=content)
        if resp.status_code >= 400:
            raise HTTPError(resp.status_code, resp.content)
        return resp.status_code, resp.content

    def request(self, method: str, url: str, headers: dict[str, str], **kwargs) -> httpx.Response:
        """Low-level request for multipart uploads etc."""
        return self._client.request(method=method, url=url, headers=headers, **kwargs)

    def close(self) -> None:
        self._client.close()


class AsyncHttpClient:
    def __init__(self, timeout_s: float = 30.0):
        self._client = httpx.AsyncClient(timeout=timeout_s)

    async def do(self, method: str, url: str, headers: dict[str, str], body: dict | None = None) -> dict | list | None:
        kwargs: dict = {"method": method, "url": url, "headers": headers}
        if body is not None:
            kwargs["json"] = body
        resp = await self._client.request(**kwargs)
        if resp.status_code >= 400:
            raise HTTPError(resp.status_code, resp.content)
        if not resp.content:
            return None
        return resp.json()

    async def do_raw(
        self,
        method: str,
        url: str,
        headers: dict[str, str],
        content: bytes | None = None,
        content_type: str | None = None,
    ) -> tuple[int, bytes]:
        h = dict(headers)
        if content_type:
            h["Content-Type"] = content_type
        resp = await self._client.request(method=method, url=url, headers=h, content=content)
        if resp.status_code >= 400:
            raise HTTPError(resp.status_code, resp.content)
        return resp.status_code, resp.content

    async def request(self, method: str, url: str, headers: dict[str, str], **kwargs) -> httpx.Response:
        return await self._client.request(method=method, url=url, headers=headers, **kwargs)

    async def close(self) -> None:
        await self._client.aclose()



from __future__ import annotations

import json
import struct
from typing import Callable, Any

from ._http_client import HttpClient, AsyncHttpClient, HTTPError

CONNECT_PROTOCOL_VERSION = "Connect-Protocol-Version"
CONNECT_TIMEOUT_MS = "Connect-Timeout-Ms"


def unary_post(hc: HttpClient, url: str, headers: dict[str, str], req_msg: Any) -> Any:
    h = {**headers, CONNECT_PROTOCOL_VERSION: "1"}
    try:
        status, raw = hc.do_raw("POST", url, h, content=json.dumps(req_msg).encode(), content_type="application/json")
    except HTTPError as e:
        raise _decode_connect_err(e) from e
    if not raw:
        return None
    parsed = json.loads(raw)
    if "error" in parsed and parsed["error"]:
        msg = parsed["error"].get("message", json.dumps(parsed["error"]))
        raise Exception(f"connect error: {msg}")
    if "result" in parsed:
        return parsed["result"]
    return parsed


def stream_post(hc: HttpClient, url: str, headers: dict[str, str], req_msg: Any, each: Callable[[Any], None]) -> None:
    h = {
        **headers,
        CONNECT_PROTOCOL_VERSION: "1",
        "Content-Type": "application/json",
        "Accept": "application/connect+json",
    }
    resp = hc.request("POST", url, h, content=json.dumps(req_msg).encode())
    if resp.status_code >= 400:
        raise _decode_connect_err(HTTPError(resp.status_code, resp.content))
    _read_connect_stream(resp.content, each)


def _read_connect_stream(data: bytes, each: Callable[[Any], None]) -> None:
    offset = 0
    while offset + 5 <= len(data):
        frame_len = struct.unpack(">I", data[offset + 1 : offset + 5])[0]
        if frame_len <= 0 or frame_len > 64 * 1024 * 1024:
            raise Exception(f"connect stream: invalid frame size {frame_len}")
        if offset + 5 + frame_len > len(data):
            break
        payload = data[offset + 5 : offset + 5 + frame_len]
        offset += 5 + frame_len
        parsed = json.loads(payload)
        if "error" in parsed and parsed["error"]:
            msg = parsed["error"].get("message", json.dumps(parsed["error"]))
            raise Exception(f"connect stream error: {msg}")
        each(parsed)


def _decode_connect_err(he: HTTPError) -> Exception:
    try:
        env = json.loads(he.body)
        if "message" in env:
            return Exception(f"http {he.status}: {env['message']}")
    except (json.JSONDecodeError, KeyError):
        pass
    return Exception(f"http {he.status}: {he.body.decode(errors='replace')}")



from __future__ import annotations

from typing import Any

from .types import *  # noqa: F403
from .errors import *  # noqa: F403
from .provider import Provider, Sandbox, AsyncProvider, AsyncSandbox
from .config import ProviderConfig
from .registry import resolve_provider, resolve_async_provider, register_provider
from ._version import VERSION

# Trigger provider registrations
from . import providers as _providers  # noqa: F401

from .types import (
    ProviderName,
    SandboxInfo,
    CreateSandboxRequest,
    ListSandboxesFilter,
)


class Sandboxer:
    """Main entry point for sync usage."""

    def __init__(self, provider: str, config: dict[str, Any] | ProviderConfig | None = None):
        if isinstance(config, dict):
            config = ProviderConfig(
                api_key=config.get("api_key") or config.get("apiKey"),
                base_url=config.get("base_url") or config.get("baseUrl"),
                default_timeout_ms=config.get("default_timeout_ms") or config.get("defaultTimeoutMs"),
                extra={
                    k: v
                    for k, v in config.items()
                    if k not in ("api_key", "apiKey", "base_url", "baseUrl", "default_timeout_ms", "defaultTimeoutMs")
                },
            )
        name = ProviderName(provider)
        self._provider = resolve_provider(name, config)

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        return self._provider.create_sandbox(req)

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        return self._provider.list_sandboxes(filter)

    def kill_sandbox(self, sandbox_id: str) -> None:
        return self._provider.kill_sandbox(sandbox_id)

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        return self._provider.attach_sandbox(sandbox_id)

    def close(self) -> None:
        return self._provider.close()


class AsyncSandboxer:
    """Main entry point for async usage."""

    def __init__(self, provider: str, config: dict[str, Any] | ProviderConfig | None = None):
        if isinstance(config, dict):
            config = ProviderConfig(
                api_key=config.get("api_key") or config.get("apiKey"),
                base_url=config.get("base_url") or config.get("baseUrl"),
                default_timeout_ms=config.get("default_timeout_ms") or config.get("defaultTimeoutMs"),
                extra={
                    k: v
                    for k, v in config.items()
                    if k not in ("api_key", "apiKey", "base_url", "baseUrl", "default_timeout_ms", "defaultTimeoutMs")
                },
            )
        name = ProviderName(provider)
        self._provider = resolve_async_provider(name, config)

    async def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[AsyncSandbox, SandboxInfo]:
        return await self._provider.create_sandbox(req)

    async def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        return await self._provider.list_sandboxes(filter)

    async def kill_sandbox(self, sandbox_id: str) -> None:
        return await self._provider.kill_sandbox(sandbox_id)

    async def attach_sandbox(self, sandbox_id: str) -> AsyncSandbox:
        return await self._provider.attach_sandbox(sandbox_id)

    async def close(self) -> None:
        return await self._provider.close()


__all__ = [
    "Sandboxer",
    "AsyncSandboxer",
    "Provider",
    "Sandbox",
    "AsyncProvider",
    "AsyncSandbox",
    "ProviderConfig",
    "ProviderName",
    "SandboxStatus",
    "SandboxInfo",
    "CommandResult",
    "FileInfo",
    "ProcessInfo",
    "PTYInfo",
    "WatchEventType",
    "WatchEvent",
    "CreateSandboxRequest",
    "RunCommandRequest",
    "StartCommandRequest",
    "CreatePTYRequest",
    "ListSandboxesFilter",
    "SandboxerError",
    "NotFoundError",
    "UnauthorizedError",
    "RateLimitError",
    "QuotaExceededError",
    "NotSupportedError",
    "BadConfigError",
    "ProviderError",
    "register_provider",
    "VERSION",
]



from __future__ import annotations

import os
import time
from datetime import datetime
from urllib.parse import quote

from ..types import (
    ProviderName,
    SandboxStatus,
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)
from ..config import ProviderConfig
from ..provider import Provider, Sandbox
from ..errors import NotFoundError, NotSupportedError, ProviderError, BadConfigError
from .._http_client import HttpClient, HTTPError
from .._util import first_non_empty, shell_quote
from ..registry import register_provider

_DEFAULT_API = "https://api.runloop.ai"


def _elapsed_ms(start: float) -> int:
    return int((time.monotonic() - start) * 1000)


def _map_err(err: HTTPError) -> Exception:
    if err.status == 404:
        return NotFoundError(str(err))
    msg = err.body.decode(errors="replace")
    if len(msg) > 512:
        msg = msg[:512] + "..."
    return ProviderError(provider="runloop", message=msg, status_code=err.status)


def _map_status(s: str) -> SandboxStatus:
    low = s.lower()
    if low in ("suspended", "suspending"):
        return SandboxStatus.PAUSED
    if low in ("shutdown", "failure"):
        return SandboxStatus.STOPPED
    if low in ("provisioning", "initializing", "resuming"):
        return SandboxStatus.STARTING
    return SandboxStatus.RUNNING


class RunloopProvider(Provider):
    def __init__(self, cfg: ProviderConfig) -> None:
        tok = first_non_empty(cfg.api_key, os.environ.get("RUNLOOP_API_KEY"))
        if not tok:
            raise BadConfigError("Runloop API key required (api_key or RUNLOOP_API_KEY)")
        timeout_s = 30.0
        if cfg.default_timeout_ms:
            timeout_s = cfg.default_timeout_ms / 1000.0
        self._hc = HttpClient(timeout_s=timeout_s)
        self._token = tok
        self._base = (cfg.base_url or _DEFAULT_API).rstrip("/")

    def _hdr(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self._token}"}

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        filter = filter or ListSandboxesFilter()
        if filter.provider is not None and filter.provider != ProviderName.RUNLOOP:
            return []
        url = f"{self._base}/v1/devboxes?limit=5000"
        try:
            page = self._hc.do("GET", url, self._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        if not page:
            return []
        devboxes = page.get("devboxes", []) if isinstance(page, dict) else []  # type: ignore[union-attr]
        out: list[SandboxInfo] = []
        for d in devboxes:
            name = d.get("name") or None
            out.append(
                SandboxInfo(
                    id=d.get("id", ""),
                    provider=ProviderName.RUNLOOP,
                    status=_map_status(d.get("status", "running")),
                    started_at=datetime.now(),
                    template=name,
                )
            )
            if filter.limit and filter.limit > 0 and len(out) >= filter.limit:
                break
        return out

    def kill_sandbox(self, sandbox_id: str) -> None:
        url = f"{self._base}/v1/devboxes/{quote(sandbox_id, safe='')}/shutdown"
        try:
            self._hc.do("POST", url, self._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        req = req or CreateSandboxRequest()
        body: dict = {}
        if req.template:
            body["blueprint_name"] = req.template
        if req.envs:
            body["environment_variables"] = req.envs
        if req.metadata:
            body["metadata"] = req.metadata
        url = f"{self._base}/v1/devboxes"
        try:
            created = self._hc.do("POST", url, self._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        cid = created.get("id", "")  # type: ignore[union-attr]
        name = created.get("name") or (req.template if req.template else None)  # type: ignore[union-attr]
        sb = RunloopSandbox(provider=self, sandbox_id=cid)
        info = SandboxInfo(
            id=cid,
            provider=ProviderName.RUNLOOP,
            status=_map_status(created.get("status", "running")),  # type: ignore[union-attr]
            started_at=datetime.now(),
            template=name,
        )
        return sb, info

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        sb = RunloopSandbox(provider=self, sandbox_id=sandbox_id)
        sb.info()  # validate
        return sb

    def close(self) -> None:
        self._hc.close()


class RunloopSandbox(Sandbox):
    def __init__(self, provider: RunloopProvider, sandbox_id: str) -> None:
        self._p = provider
        self._id = sandbox_id

    @property
    def id(self) -> str:
        return self._id

    def info(self) -> SandboxInfo:
        url = f"{self._p._base}/v1/devboxes/{quote(self._id, safe='')}"
        try:
            d = self._p._hc.do("GET", url, self._p._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        name = d.get("name") or None  # type: ignore[union-attr]
        return SandboxInfo(
            id=d.get("id", ""),  # type: ignore[union-attr]
            provider=ProviderName.RUNLOOP,
            status=_map_status(d.get("status", "running")),  # type: ignore[union-attr]
            started_at=datetime.now(),
            template=name,
        )

    def is_running(self) -> bool:
        return self.info().status == SandboxStatus.RUNNING

    def pause(self) -> None:
        url = f"{self._p._base}/v1/devboxes/{quote(self._id, safe='')}/suspend"
        try:
            self._p._hc.do("POST", url, self._p._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def resume(self) -> None:
        url = f"{self._p._base}/v1/devboxes/{quote(self._id, safe='')}/resume"
        try:
            self._p._hc.do("POST", url, self._p._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def kill(self) -> None:
        self._p.kill_sandbox(self._id)

    def port_url(self, port: int) -> str:
        raise NotSupportedError("Runloop does not support port_url")

    def run_command(self, req: RunCommandRequest) -> CommandResult:
        start = time.monotonic()
        url = f"{self._p._base}/v1/devboxes/{quote(self._id, safe='')}/execute_sync"
        body: dict = {"command": req.cmd}
        shell_name = os.environ.get("RUNLOOP_SHELL_NAME")
        if shell_name:
            body["shell_name"] = shell_name
        try:
            out = self._p._hc.do("POST", url, self._p._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        if not out:
            out = {}
        return CommandResult(
            stdout=out.get("stdout", ""),  # type: ignore[union-attr]
            stderr=out.get("stderr", ""),  # type: ignore[union-attr]
            exit_code=out.get("exit_status", 0),  # type: ignore[union-attr]
            duration_ms=_elapsed_ms(start),
        )

    def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        raise NotSupportedError("Runloop does not support start_command")

    def wait_for_handle(self, handle_id: str) -> CommandResult:
        raise NotSupportedError("Runloop does not support wait_for_handle")

    def kill_process(self, pid: int) -> None:
        raise NotSupportedError("Runloop does not support kill_process")

    def list_processes(self) -> list[ProcessInfo]:
        raise NotSupportedError("Runloop does not support list_processes")

    def read_file(self, path: str) -> bytes:
        url = f"{self._p._base}/v1/devboxes/{quote(self._id, safe='')}/read_file_contents"
        body = {"file_path": path}
        try:
            out = self._p._hc.do("POST", url, self._p._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        contents = out.get("contents", "") if isinstance(out, dict) else ""
        return contents.encode()

    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        url = f"{self._p._base}/v1/devboxes/{quote(self._id, safe='')}/write_file_contents"
        body = {"file_path": path, "contents": content.decode(errors="replace")}
        try:
            self._p._hc.do("POST", url, self._p._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e

    def list_directory(self, path: str) -> list[FileInfo]:
        res = self.run_command(RunCommandRequest(cmd="ls -1 " + shell_quote(path)))
        if res.exit_code != 0:
            raise Exception(f"ls failed: {res.stderr}")
        out: list[FileInfo] = []
        for line in res.stdout.strip().split("\n"):
            line = line.strip()
            if not line:
                continue
            p = path.rstrip("/") + "/" + line
            out.append(FileInfo(name=line, path=p))
        return out

    def make_dir(self, path: str) -> None:
        self.run_command(RunCommandRequest(cmd="mkdir -p " + shell_quote(path)))

    def remove(self, path: str) -> None:
        self.run_command(RunCommandRequest(cmd="rm -rf " + shell_quote(path)))

    def exists(self, path: str) -> bool:
        res = self.run_command(RunCommandRequest(cmd="test -e " + shell_quote(path) + " && echo yes || echo no"))
        return res.stdout.strip() == "yes"

    def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        raise NotSupportedError("Runloop does not support create_pty")

    def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        raise NotSupportedError("Runloop does not support resize_pty")

    def kill_pty(self, pid: int) -> None:
        raise NotSupportedError("Runloop does not support kill_pty")

    def list_pty(self) -> list[PTYInfo]:
        raise NotSupportedError("Runloop does not support list_pty")


def _make_runloop_provider(cfg: ProviderConfig) -> Provider:
    return RunloopProvider(cfg)


register_provider(ProviderName.RUNLOOP, _make_runloop_provider)



from __future__ import annotations


class SandboxerError(Exception):
    pass


class NotFoundError(SandboxerError):
    pass


class UnauthorizedError(SandboxerError):
    pass


class RateLimitError(SandboxerError):
    pass


class QuotaExceededError(SandboxerError):
    pass


class NotSupportedError(SandboxerError):
    pass


class BadConfigError(SandboxerError):
    pass


class ProviderError(SandboxerError):
    def __init__(
        self,
        provider: str,
        message: str,
        status_code: int | None = None,
        code: str | None = None,
    ):
        super().__init__(f"provider {provider}: {message}")
        self.provider = provider
        self.status_code = status_code
        self.code = code



from __future__ import annotations

from abc import ABC, abstractmethod

from .types import (
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)


class Sandbox(ABC):
    @property
    @abstractmethod
    def id(self) -> str: ...

    @abstractmethod
    def info(self) -> SandboxInfo: ...

    @abstractmethod
    def is_running(self) -> bool: ...

    @abstractmethod
    def pause(self) -> None: ...

    @abstractmethod
    def resume(self) -> None: ...

    @abstractmethod
    def kill(self) -> None: ...

    @abstractmethod
    def port_url(self, port: int) -> str: ...

    @abstractmethod
    def run_command(self, req: RunCommandRequest) -> CommandResult: ...

    @abstractmethod
    def start_command(self, req: StartCommandRequest) -> tuple[int, str]: ...

    @abstractmethod
    def wait_for_handle(self, handle_id: str) -> CommandResult: ...

    @abstractmethod
    def kill_process(self, pid: int) -> None: ...

    @abstractmethod
    def list_processes(self) -> list[ProcessInfo]: ...

    @abstractmethod
    def read_file(self, path: str) -> bytes: ...

    @abstractmethod
    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None: ...

    @abstractmethod
    def list_directory(self, path: str) -> list[FileInfo]: ...

    @abstractmethod
    def make_dir(self, path: str) -> None: ...

    @abstractmethod
    def remove(self, path: str) -> None: ...

    @abstractmethod
    def exists(self, path: str) -> bool: ...

    @abstractmethod
    def create_pty(self, req: CreatePTYRequest) -> PTYInfo: ...

    @abstractmethod
    def resize_pty(self, pid: int, rows: int, cols: int) -> None: ...

    @abstractmethod
    def kill_pty(self, pid: int) -> None: ...

    @abstractmethod
    def list_pty(self) -> list[PTYInfo]: ...


class Provider(ABC):
    @abstractmethod
    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]: ...

    @abstractmethod
    def kill_sandbox(self, sandbox_id: str) -> None: ...

    @abstractmethod
    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]: ...

    @abstractmethod
    def attach_sandbox(self, sandbox_id: str) -> Sandbox: ...

    @abstractmethod
    def close(self) -> None: ...


class AsyncSandbox(ABC):
    @property
    @abstractmethod
    def id(self) -> str: ...

    @abstractmethod
    async def info(self) -> SandboxInfo: ...

    @abstractmethod
    async def is_running(self) -> bool: ...

    @abstractmethod
    async def pause(self) -> None: ...

    @abstractmethod
    async def resume(self) -> None: ...

    @abstractmethod
    async def kill(self) -> None: ...

    @abstractmethod
    async def port_url(self, port: int) -> str: ...

    @abstractmethod
    async def run_command(self, req: RunCommandRequest) -> CommandResult: ...

    @abstractmethod
    async def start_command(self, req: StartCommandRequest) -> tuple[int, str]: ...

    @abstractmethod
    async def wait_for_handle(self, handle_id: str) -> CommandResult: ...

    @abstractmethod
    async def kill_process(self, pid: int) -> None: ...

    @abstractmethod
    async def list_processes(self) -> list[ProcessInfo]: ...

    @abstractmethod
    async def read_file(self, path: str) -> bytes: ...

    @abstractmethod
    async def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None: ...

    @abstractmethod
    async def list_directory(self, path: str) -> list[FileInfo]: ...

    @abstractmethod
    async def make_dir(self, path: str) -> None: ...

    @abstractmethod
    async def remove(self, path: str) -> None: ...

    @abstractmethod
    async def exists(self, path: str) -> bool: ...

    @abstractmethod
    async def create_pty(self, req: CreatePTYRequest) -> PTYInfo: ...

    @abstractmethod
    async def resize_pty(self, pid: int, rows: int, cols: int) -> None: ...

    @abstractmethod
    async def kill_pty(self, pid: int) -> None: ...

    @abstractmethod
    async def list_pty(self) -> list[PTYInfo]: ...


class AsyncProvider(ABC):
    @abstractmethod
    async def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]: ...

    @abstractmethod
    async def kill_sandbox(self, sandbox_id: str) -> None: ...

    @abstractmethod
    async def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[AsyncSandbox, SandboxInfo]: ...

    @abstractmethod
    async def attach_sandbox(self, sandbox_id: str) -> AsyncSandbox: ...

    @abstractmethod
    async def close(self) -> None: ...



from __future__ import annotations

import io
import os
import time
from datetime import datetime
from urllib.parse import quote, urlencode

from ..types import (
    ProviderName,
    SandboxStatus,
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)
from ..config import ProviderConfig
from ..provider import Provider, Sandbox
from ..errors import NotFoundError, NotSupportedError, ProviderError, BadConfigError
from .._http_client import HttpClient, HTTPError
from .._util import first_non_empty
from ..registry import register_provider

_DEFAULT_API = "https://app.daytona.io/api"
_DEFAULT_TOOLBOX = "https://proxy.app.daytona.io/toolbox"


def _elapsed_ms(start: float) -> int:
    return int((time.monotonic() - start) * 1000)


def _map_err(err: HTTPError) -> Exception:
    if err.status == 404:
        return NotFoundError(str(err))
    msg = err.body.decode(errors="replace")
    return ProviderError(provider="daytona", message=msg, status_code=err.status)


def _map_state(state: str) -> SandboxStatus:
    s = state.lower()
    if s in ("stopped", "archived"):
        return SandboxStatus.STOPPED
    if s in ("starting", "creating"):
        return SandboxStatus.STARTING
    return SandboxStatus.RUNNING


class DaytonaProvider(Provider):
    def __init__(self, cfg: ProviderConfig) -> None:
        tok = first_non_empty(cfg.api_key, os.environ.get("DAYTONA_API_KEY"), os.environ.get("DAYTONA_TOKEN"))
        if not tok:
            raise BadConfigError("Daytona API token required (api_key, DAYTONA_API_KEY, or DAYTONA_TOKEN)")
        timeout_s = 30.0
        if cfg.default_timeout_ms:
            timeout_s = cfg.default_timeout_ms / 1000.0
        self._hc = HttpClient(timeout_s=timeout_s)
        self._token = tok
        self._api_base = (cfg.base_url or _DEFAULT_API).rstrip("/")
        self._tool_base = (os.environ.get("DAYTONA_TOOLBOX_BASE_URL") or _DEFAULT_TOOLBOX).rstrip("/")

    def _hdr(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self._token}"}

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        filter = filter or ListSandboxesFilter()
        if filter.provider is not None and filter.provider != ProviderName.DAYTONA:
            return []
        url = f"{self._api_base}/sandbox"
        if filter.limit and filter.limit > 0:
            url += f"?limit={filter.limit}"
        try:
            rows = self._hc.do("GET", url, self._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        if not rows:
            return []
        out: list[SandboxInfo] = []
        for s in rows:  # type: ignore[union-attr]
            metadata = s.get("labels") or {}
            if filter.metadata_filter and not any(filter.metadata_filter in v for v in metadata.values()):
                continue
            sid = first_non_empty(s.get("id"), s.get("name"))
            info = SandboxInfo(
                id=sid,
                provider=ProviderName.DAYTONA,
                status=_map_state(s.get("state", "running")),
                started_at=datetime.now(),
                metadata=metadata,
                template=s.get("image") or None,
            )
            out.append(info)
        return out

    def kill_sandbox(self, sandbox_id: str) -> None:
        url = f"{self._api_base}/sandbox/{quote(sandbox_id, safe='')}"
        try:
            self._hc.do("DELETE", url, self._hdr())
        except HTTPError as e:
            raise _map_err(e) from e

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        req = req or CreateSandboxRequest()
        body: dict = {"env": req.metadata}
        if req.template:
            body["image"] = req.template
        if req.envs:
            body["envVars"] = req.envs
        if req.cpus is not None or req.memory_mb is not None:
            res: dict = {}
            if req.cpus is not None:
                res["cpu"] = req.cpus
            if req.memory_mb is not None:
                res["memory"] = req.memory_mb
            body["resources"] = res
        url = f"{self._api_base}/sandbox"
        try:
            created = self._hc.do("POST", url, self._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        sid = first_non_empty(created.get("id"), created.get("name"))  # type: ignore[union-attr]
        sb = DaytonaSandbox(provider=self, sandbox_id=sid)
        try:
            info = sb.info()
        except Exception:
            info = SandboxInfo(
                id=sid,
                provider=ProviderName.DAYTONA,
                status=SandboxStatus.RUNNING,
                started_at=datetime.now(),
                template=created.get("image") or None,  # type: ignore[union-attr]
            )
        return sb, info

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        sb = DaytonaSandbox(provider=self, sandbox_id=sandbox_id)
        sb.info()  # validate existence
        return sb

    def close(self) -> None:
        self._hc.close()


class DaytonaSandbox(Sandbox):
    def __init__(self, provider: DaytonaProvider, sandbox_id: str) -> None:
        self._p = provider
        self._id = sandbox_id

    @property
    def id(self) -> str:
        return self._id

    def info(self) -> SandboxInfo:
        url = f"{self._p._api_base}/sandbox/{quote(self._id, safe='')}"
        try:
            d = self._p._hc.do("GET", url, self._p._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        sid = first_non_empty(d.get("id"), d.get("name"))  # type: ignore[union-attr]
        return SandboxInfo(
            id=sid,
            provider=ProviderName.DAYTONA,
            status=_map_state(d.get("state", "running")),  # type: ignore[union-attr]
            started_at=datetime.now(),
            template=d.get("image") or None,  # type: ignore[union-attr]
        )

    def is_running(self) -> bool:
        i = self.info()
        return i.status in (SandboxStatus.RUNNING, SandboxStatus.STARTING)

    def pause(self) -> None:
        url = f"{self._p._api_base}/sandbox/{quote(self._id, safe='')}/stop"
        try:
            self._p._hc.do("POST", url, self._p._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def resume(self) -> None:
        url = f"{self._p._api_base}/sandbox/{quote(self._id, safe='')}/start"
        try:
            self._p._hc.do("POST", url, self._p._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def kill(self) -> None:
        self._p.kill_sandbox(self._id)

    def port_url(self, port: int) -> str:
        raise NotSupportedError("Daytona does not support port_url")

    def run_command(self, req: RunCommandRequest) -> CommandResult:
        start = time.monotonic()
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/process/execute"
        body: dict = {"command": req.cmd}
        if req.cwd:
            body["cwd"] = req.cwd
        if req.timeout_seconds is not None:
            body["timeout"] = req.timeout_seconds
        if req.env:
            body["env"] = req.env
        try:
            out = self._p._hc.do("POST", url, self._p._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        if not out:
            out = {}
        code = out.get("exitCode", out.get("exit_code", 0))  # type: ignore[union-attr]
        stdout = first_non_empty(out.get("stdout"), out.get("result"))  # type: ignore[union-attr]
        return CommandResult(
            stdout=stdout,
            stderr=out.get("stderr", ""),  # type: ignore[union-attr]
            exit_code=code,
            duration_ms=_elapsed_ms(start),
        )

    def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        raise NotSupportedError("Daytona does not support start_command")

    def wait_for_handle(self, handle_id: str) -> CommandResult:
        raise NotSupportedError("Daytona does not support wait_for_handle")

    def kill_process(self, pid: int) -> None:
        raise NotSupportedError("Daytona does not support kill_process")

    def list_processes(self) -> list[ProcessInfo]:
        raise NotSupportedError("Daytona does not support list_processes")

    def read_file(self, path: str) -> bytes:
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/files/download?{urlencode({'path': path})}"
        try:
            _, data = self._p._hc.do_raw("GET", url, self._p._hdr())
        except HTTPError as e:
            if e.status == 404:
                raise NotFoundError(f"file not found: {path}") from e
            raise _map_err(e) from e
        return data

    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/files/upload?{urlencode({'path': path})}"
        h = self._p._hdr()
        boundary = "----SandboxerBoundary"
        body = io.BytesIO()
        body.write(f"--{boundary}\r\n".encode())
        body.write(b'Content-Disposition: form-data; name="path"\r\n\r\n')
        body.write(path.encode())
        body.write(b"\r\n")
        body.write(f"--{boundary}\r\n".encode())
        body.write(b'Content-Disposition: form-data; name="file"; filename="file"\r\n')
        body.write(b"Content-Type: application/octet-stream\r\n\r\n")
        body.write(content)
        body.write(b"\r\n")
        body.write(f"--{boundary}--\r\n".encode())
        h["Content-Type"] = f"multipart/form-data; boundary={boundary}"
        try:
            self._p._hc.do_raw("POST", url, h, content=body.getvalue())
        except HTTPError as e:
            raise _map_err(e) from e

    def list_directory(self, path: str) -> list[FileInfo]:
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/files?{urlencode({'path': path})}"
        try:
            _, raw = self._p._hc.do_raw("GET", url, self._p._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        import json

        try:
            entries = json.loads(raw)
        except json.JSONDecodeError:
            return []
        # Could be a list or a wrapper with "entries" key
        if isinstance(entries, dict):
            entries = entries.get("entries", [])
        out: list[FileInfo] = []
        for e in entries:
            p = e.get("path") or (path.rstrip("/") + "/" + e.get("name", ""))
            out.append(
                FileInfo(
                    name=e.get("name", ""),
                    path=p,
                    is_dir=e.get("isDir", False),
                    size=e.get("size", 0),
                )
            )
        return out

    def make_dir(self, path: str) -> None:
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/files/folder?{urlencode({'path': path, 'mode': '755'})}"
        try:
            self._p._hc.do_raw("POST", url, self._p._hdr())
        except HTTPError as e:
            raise _map_err(e) from e

    def remove(self, path: str) -> None:
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/files?{urlencode({'path': path})}"
        try:
            self._p._hc.do_raw("DELETE", url, self._p._hdr())
        except HTTPError as e:
            raise _map_err(e) from e

    def exists(self, path: str) -> bool:
        url = f"{self._p._tool_base}/{quote(self._id, safe='')}/files/info?{urlencode({'path': path})}"
        try:
            self._p._hc.do_raw("GET", url, self._p._hdr())
        except HTTPError as e:
            if e.status == 404:
                return False
            raise _map_err(e) from e
        return True

    def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        raise NotSupportedError("Daytona does not support create_pty")

    def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        raise NotSupportedError("Daytona does not support resize_pty")

    def kill_pty(self, pid: int) -> None:
        raise NotSupportedError("Daytona does not support kill_pty")

    def list_pty(self) -> list[PTYInfo]:
        raise NotSupportedError("Daytona does not support list_pty")


def _make_daytona_provider(cfg: ProviderConfig) -> Provider:
    return DaytonaProvider(cfg)


register_provider(ProviderName.DAYTONA, _make_daytona_provider)



from __future__ import annotations

from enum import StrEnum
from datetime import datetime

from pydantic import BaseModel, Field


class ProviderName(StrEnum):
    E2B = "e2b"
    DAYTONA = "daytona"
    BLAXEL = "blaxel"
    RUNLOOP = "runloop"
    FLY_MACHINES = "fly-machines"
    LOCAL = "local"


class SandboxStatus(StrEnum):
    STARTING = "starting"
    RUNNING = "running"
    PAUSED = "paused"
    STOPPED = "stopped"
    ERROR = "error"


class SandboxInfo(BaseModel):
    id: str
    provider: ProviderName
    template: str | None = None
    status: SandboxStatus
    started_at: datetime = Field(default_factory=datetime.now)
    expires_at: datetime | None = None
    metadata: dict[str, str] = Field(default_factory=dict)
    cpus: int | None = None
    memory_mb: int | None = None


class CommandResult(BaseModel):
    stdout: str = ""
    stderr: str = ""
    exit_code: int = 0
    duration_ms: int = 0
    error: str | None = None


class FileInfo(BaseModel):
    name: str
    path: str
    is_dir: bool = False
    size: int = 0
    mode: int | None = None


class ProcessInfo(BaseModel):
    pid: int
    command: str
    user: str | None = None


class PTYInfo(BaseModel):
    pid: int
    rows: int
    cols: int


class WatchEventType(StrEnum):
    CREATE = "create"
    MODIFY = "modify"
    DELETE = "delete"
    RENAME = "rename"


class WatchEvent(BaseModel):
    path: str
    event_type: WatchEventType


class CreateSandboxRequest(BaseModel):
    provider: ProviderName | None = None
    template: str | None = None
    timeout_seconds: int | None = None
    metadata: dict[str, str] = Field(default_factory=dict)
    envs: dict[str, str] = Field(default_factory=dict)
    cpus: int | None = None
    memory_mb: int | None = None
    auto_destroy: bool | None = None


class RunCommandRequest(BaseModel):
    cmd: str
    cwd: str | None = None
    env: dict[str, str] = Field(default_factory=dict)
    timeout_seconds: int | None = None
    user: str | None = None


class StartCommandRequest(BaseModel):
    cmd: str
    cwd: str | None = None
    env: dict[str, str] = Field(default_factory=dict)
    user: str | None = None


class CreatePTYRequest(BaseModel):
    rows: int | None = None
    cols: int | None = None
    cwd: str | None = None
    env: dict[str, str] = Field(default_factory=dict)
    user: str | None = None
    command: str | None = None


class ListSandboxesFilter(BaseModel):
    provider: ProviderName | None = None
    metadata_filter: str | None = None
    limit: int | None = None



from __future__ import annotations

import asyncio
import json
import os
import subprocess
import tarfile
import threading
import time
from datetime import datetime
from io import BytesIO
from pathlib import PurePosixPath

from ..types import (
    ProviderName,
    SandboxStatus,
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)
from ..config import ProviderConfig
from ..provider import Provider, Sandbox, AsyncProvider, AsyncSandbox
from ..errors import NotFoundError, NotSupportedError, ProviderError, BadConfigError
from .._util import normalize_path
from ..registry import register_provider

_LABEL_MANAGED = "sandboxer.managed"
_LABEL_PROVIDER = "sandboxer.provider"


def _elapsed_ms(start: float) -> int:
    return int((time.monotonic() - start) * 1000)


def _docker_run(args: list[str], input_data: bytes | None = None) -> bytes:
    cmd = ["docker"] + args
    result = subprocess.run(cmd, capture_output=True, input=input_data)
    if result.returncode != 0:
        stderr = result.stderr.decode(errors="replace").strip()
        raise ProviderError(provider="local", message=f"docker {' '.join(args[:2])}: {stderr}")
    return result.stdout


def _sanitize_label_key(k: str) -> str:
    out = []
    for ch in k:
        if ch.isalnum() or ch in (".", "_", "-"):
            out.append(ch)
        else:
            out.append("_")
    return "".join(out) or "key"


def _metadata_contains(m: dict[str, str], needle: str) -> bool:
    if not needle:
        return True
    for v in m.values():
        if needle in v:
            return True
    return False


class LocalProvider(Provider):
    def __init__(self, cfg: ProviderConfig) -> None:
        import shutil

        if not shutil.which("docker"):
            raise BadConfigError("docker CLI not found in PATH")
        # Validate docker is working
        try:
            subprocess.run(["docker", "info"], capture_output=True, check=True)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            raise BadConfigError(f"docker info failed: {e}") from e
        self._cfg = cfg

    def _inspect(self, container_id: str) -> SandboxInfo:
        raw = _docker_run(["inspect", container_id])
        try:
            wrap = json.loads(raw)
        except json.JSONDecodeError as e:
            raise ProviderError(provider="local", message=f"inspect decode: {e}") from e
        if not wrap:
            raise NotFoundError(f"container {container_id} not found")
        c = wrap[0]
        labels = c.get("Config", {}).get("Labels", {}) or {}
        if labels.get(_LABEL_MANAGED) != "true":
            raise NotFoundError(f"container {container_id} not managed by sandboxer")
        state = c.get("State", {})
        if state.get("Running"):
            st = SandboxStatus.RUNNING
        elif state.get("Paused"):
            st = SandboxStatus.PAUSED
        elif state.get("Status") in ("created", "restarting"):
            st = SandboxStatus.STARTING
        elif state.get("OOMKilled") or state.get("ExitCode", 0) != 0:
            st = SandboxStatus.ERROR
        else:
            st = SandboxStatus.STOPPED
        meta: dict[str, str] = {}
        for k, v in labels.items():
            if k.startswith("sandboxer.meta."):
                meta[k.removeprefix("sandboxer.meta.")] = v
        started = datetime.now()
        if state.get("StartedAt"):
            try:
                started = datetime.fromisoformat(state["StartedAt"].replace("Z", "+00:00"))
            except (ValueError, TypeError):
                pass
        image = c.get("Config", {}).get("Image", "")
        return SandboxInfo(
            id=c.get("Id", container_id),
            provider=ProviderName.LOCAL,
            template=image or None,
            status=st,
            started_at=started,
            metadata=meta,
        )

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        filter = filter or ListSandboxesFilter()
        if filter.provider is not None and filter.provider != ProviderName.LOCAL:
            return []
        raw = _docker_run(["ps", "-a", "-q", "--no-trunc", "--filter", f"label={_LABEL_MANAGED}=true"])
        ids = raw.decode().strip().split()
        out: list[SandboxInfo] = []
        for cid in ids:
            if not cid:
                continue
            try:
                info = self._inspect(cid)
            except Exception:
                continue
            if filter.metadata_filter and not _metadata_contains(info.metadata, filter.metadata_filter):
                continue
            out.append(info)
            if filter.limit and filter.limit > 0 and len(out) >= filter.limit:
                break
        return out

    def kill_sandbox(self, sandbox_id: str) -> None:
        try:
            _docker_run(["rm", "-f", sandbox_id])
        except ProviderError:
            pass  # already gone

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        req = req or CreateSandboxRequest()
        image = req.template or "alpine:latest"
        args = ["create", "--label", f"{_LABEL_MANAGED}=true", "--label", f"{_LABEL_PROVIDER}=local"]
        for k, v in req.metadata.items():
            args.extend(["--label", f"sandboxer.meta.{_sanitize_label_key(k)}={v}"])
        for k, v in req.envs.items():
            args.extend(["-e", f"{k}={v}"])
        if req.cpus is not None and req.cpus > 0:
            args.extend(["--cpus", str(req.cpus)])
        if req.memory_mb is not None and req.memory_mb > 0:
            args.extend(["-m", f"{req.memory_mb}m"])
        args.extend([image, "sleep", "infinity"])
        out = _docker_run(args)
        cid = out.decode().strip()
        try:
            _docker_run(["start", cid])
        except ProviderError:
            _docker_run(["rm", "-f", cid])
            raise
        info = self._inspect(cid)
        return LocalSandbox(cid), info

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        info = self._inspect(sandbox_id)
        if info.status != SandboxStatus.RUNNING:
            raise NotFoundError(f"container {sandbox_id} is not running")
        return LocalSandbox(sandbox_id)

    def close(self) -> None:
        pass


class LocalSandbox(Sandbox):
    def __init__(self, container_id: str) -> None:
        self._id = container_id.strip()
        self._handle_seq = 0
        self._handle_lock = threading.Lock()
        self._handles: dict[str, threading.Event] = {}
        self._handle_results: dict[str, tuple[CommandResult, Exception | None]] = {}

    @property
    def id(self) -> str:
        return self._id

    def info(self) -> SandboxInfo:
        p = LocalProvider.__new__(LocalProvider)
        p._cfg = ProviderConfig()
        return p._inspect(self._id)

    def is_running(self) -> bool:
        return self.info().status == SandboxStatus.RUNNING

    def pause(self) -> None:
        _docker_run(["pause", self._id])

    def resume(self) -> None:
        _docker_run(["unpause", self._id])

    def kill(self) -> None:
        _docker_run(["rm", "-f", self._id])

    def port_url(self, port: int) -> str:
        raw = _docker_run(["inspect", self._id])
        wrap = json.loads(raw)
        if not wrap:
            raise ProviderError(provider="local", message="inspect failed")
        ports = wrap[0].get("NetworkSettings", {}).get("Ports", {})
        key = f"{port}/tcp"
        binds = ports.get(key, [])
        if not binds:
            raise NotSupportedError(f"port {port} not exposed")
        host_ip = binds[0].get("HostIp", "") or "127.0.0.1"
        if host_ip == "0.0.0.0":
            host_ip = "127.0.0.1"
        return f"http://{host_ip}:{binds[0]['HostPort']}"

    def run_command(self, req: RunCommandRequest) -> CommandResult:
        start = time.monotonic()
        args = ["exec", "-i"]
        if req.user:
            args.extend(["-u", req.user])
        if req.cwd:
            args.extend(["-w", req.cwd])
        args.extend([self._id, "/bin/sh", "-c", req.cmd])
        cmd = ["docker"] + args
        timeout = None
        if req.timeout_seconds is not None and req.timeout_seconds > 0:
            timeout = req.timeout_seconds
        env = None
        if req.env:
            env = {**os.environ, **req.env}
        try:
            result = subprocess.run(cmd, capture_output=True, timeout=timeout, env=env)
        except subprocess.TimeoutExpired:
            return CommandResult(
                stdout="",
                stderr="command timed out",
                exit_code=-1,
                duration_ms=_elapsed_ms(start),
            )
        except Exception as e:
            if "No such container" in str(e):
                raise NotFoundError(f"container {self._id} not found") from e
            raise ProviderError(provider="local", message=str(e)) from e
        return CommandResult(
            stdout=result.stdout.decode(errors="replace"),
            stderr=result.stderr.decode(errors="replace"),
            exit_code=result.returncode,
            duration_ms=_elapsed_ms(start),
        )

    def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        with self._handle_lock:
            self._handle_seq += 1
            n = self._handle_seq
        handle_id = f"h{time.time_ns()}-{n}"
        event = threading.Event()
        self._handles[handle_id] = event

        def _run() -> None:
            try:
                res = self.run_command(RunCommandRequest(cmd=req.cmd, cwd=req.cwd, env=req.env, user=req.user))
                self._handle_results[handle_id] = (res, None)
            except Exception as e:
                self._handle_results[handle_id] = (CommandResult(), e)
            event.set()

        t = threading.Thread(target=_run, daemon=True)
        t.start()
        return n, handle_id

    def wait_for_handle(self, handle_id: str) -> CommandResult:
        event = self._handles.pop(handle_id, None)
        if event is None:
            raise NotFoundError(f"handle {handle_id} not found")
        event.wait()
        res, err = self._handle_results.pop(handle_id, (CommandResult(), None))
        if err:
            raise err
        return res

    def kill_process(self, pid: int) -> None:
        self.run_command(RunCommandRequest(cmd=f"kill -9 {pid} 2>/dev/null || true"))

    def list_processes(self) -> list[ProcessInfo]:
        raw = _docker_run(["top", self._id, "-eo", "pid,args"])
        lines = raw.decode(errors="replace").strip().split("\n")
        if len(lines) < 2:
            return []
        out: list[ProcessInfo] = []
        for line in lines[1:]:
            fields = line.split(None, 1)
            if len(fields) < 2:
                continue
            try:
                pid = int(fields[0])
            except ValueError:
                continue
            out.append(ProcessInfo(pid=pid, command=fields[1]))
        return out

    def read_file(self, path: str) -> bytes:
        path = normalize_path(path)
        cmd = ["docker", "cp", f"{self._id}:{path}", "-"]
        result = subprocess.run(cmd, capture_output=True)
        if result.returncode != 0:
            stderr = result.stderr.decode(errors="replace")
            if "Could not find the file" in stderr or "No such container" in stderr:
                raise NotFoundError(f"file not found: {path}")
            raise ProviderError(provider="local", message=f"docker cp: {stderr.strip()}")
        tf = tarfile.open(fileobj=BytesIO(result.stdout), mode="r:")
        member = tf.next()
        if member is None:
            raise NotFoundError(f"file not found in archive: {path}")
        f = tf.extractfile(member)
        if f is None:
            raise NotFoundError(f"cannot extract: {path}")
        return f.read()

    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        path = normalize_path(path)
        pp = PurePosixPath(path)
        dir_part = str(pp.parent) or "/"
        base = pp.name
        buf = BytesIO()
        with tarfile.open(fileobj=buf, mode="w:") as tw:
            info = tarfile.TarInfo(name=base)
            info.size = len(content)
            info.mode = mode if mode is not None else 0o644
            tw.addfile(info, BytesIO(content))
        cmd = ["docker", "cp", "-", f"{self._id}:{dir_part}"]
        result = subprocess.run(cmd, input=buf.getvalue(), capture_output=True)
        if result.returncode != 0:
            stderr = result.stderr.decode(errors="replace").strip()
            raise ProviderError(provider="local", message=f"docker cp: {stderr}")

    def list_directory(self, path: str) -> list[FileInfo]:
        path = normalize_path(path)
        res = self.run_command(RunCommandRequest(cmd=f'ls -1b "{path}"'))
        if res.exit_code != 0:
            raise NotFoundError(f"directory not found: {path}")
        names = res.stdout.strip().split("\n")
        out: list[FileInfo] = []
        for name in names:
            name = name.strip()
            if not name:
                continue
            full = str(PurePosixPath(path) / name)
            # stat for size/mode/isdir
            stat_res = self.run_command(RunCommandRequest(cmd=f"stat -c '%s %f' \"{full}\" 2>/dev/null"))
            if stat_res.exit_code != 0:
                out.append(FileInfo(name=name, path=full))
                continue
            fields = stat_res.stdout.strip().split()
            if len(fields) < 2:
                out.append(FileInfo(name=name, path=full))
                continue
            try:
                size = int(fields[0])
            except ValueError:
                size = 0
            try:
                mode_hex = int(fields[1], 16)
            except ValueError:
                mode_hex = 0
            file_mode = mode_hex & 0o7777
            is_dir = (mode_hex & 0o40000) != 0
            out.append(FileInfo(name=name, path=full, is_dir=is_dir, size=size, mode=file_mode))
        return out

    def make_dir(self, path: str) -> None:
        path = normalize_path(path)
        self.run_command(RunCommandRequest(cmd=f'mkdir -p "{path}"'))

    def remove(self, path: str) -> None:
        path = normalize_path(path)
        self.run_command(RunCommandRequest(cmd=f'rm -rf "{path}"'))

    def exists(self, path: str) -> bool:
        path = normalize_path(path)
        res = self.run_command(RunCommandRequest(cmd=f'test -e "{path}" && echo ok'))
        return res.exit_code == 0 and "ok" in res.stdout

    def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        raise NotSupportedError("Local does not support create_pty")

    def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        raise NotSupportedError("Local does not support resize_pty")

    def kill_pty(self, pid: int) -> None:
        raise NotSupportedError("Local does not support kill_pty")

    def list_pty(self) -> list[PTYInfo]:
        raise NotSupportedError("Local does not support list_pty")


class AsyncLocalSandbox(AsyncSandbox):
    """Async wrapper around `LocalSandbox` (uses a thread pool for I/O)."""

    def __init__(self, inner: LocalSandbox) -> None:
        self._s = inner

    @property
    def id(self) -> str:
        return self._s.id

    async def info(self) -> SandboxInfo:
        return await asyncio.to_thread(self._s.info)

    async def is_running(self) -> bool:
        return await asyncio.to_thread(self._s.is_running)

    async def pause(self) -> None:
        return await asyncio.to_thread(self._s.pause)

    async def resume(self) -> None:
        return await asyncio.to_thread(self._s.resume)

    async def kill(self) -> None:
        return await asyncio.to_thread(self._s.kill)

    async def port_url(self, port: int) -> str:
        return await asyncio.to_thread(self._s.port_url, port)

    async def run_command(self, req: RunCommandRequest) -> CommandResult:
        return await asyncio.to_thread(self._s.run_command, req)

    async def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        return await asyncio.to_thread(self._s.start_command, req)

    async def wait_for_handle(self, handle_id: str) -> CommandResult:
        return await asyncio.to_thread(self._s.wait_for_handle, handle_id)

    async def kill_process(self, pid: int) -> None:
        return await asyncio.to_thread(self._s.kill_process, pid)

    async def list_processes(self) -> list[ProcessInfo]:
        return await asyncio.to_thread(self._s.list_processes)

    async def read_file(self, path: str) -> bytes:
        return await asyncio.to_thread(self._s.read_file, path)

    async def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        return await asyncio.to_thread(self._s.write_file, path, content, mode, user)

    async def list_directory(self, path: str) -> list[FileInfo]:
        return await asyncio.to_thread(self._s.list_directory, path)

    async def make_dir(self, path: str) -> None:
        return await asyncio.to_thread(self._s.make_dir, path)

    async def remove(self, path: str) -> None:
        return await asyncio.to_thread(self._s.remove, path)

    async def exists(self, path: str) -> bool:
        return await asyncio.to_thread(self._s.exists, path)

    async def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        return await asyncio.to_thread(self._s.create_pty, req)

    async def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        return await asyncio.to_thread(self._s.resize_pty, pid, rows, cols)

    async def kill_pty(self, pid: int) -> None:
        return await asyncio.to_thread(self._s.kill_pty, pid)

    async def list_pty(self) -> list[PTYInfo]:
        return await asyncio.to_thread(self._s.list_pty)


class AsyncLocalProvider(AsyncProvider):
    def __init__(self, cfg: ProviderConfig) -> None:
        self._p = LocalProvider(cfg)

    async def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        return await asyncio.to_thread(self._p.list_sandboxes, filter)

    async def kill_sandbox(self, sandbox_id: str) -> None:
        return await asyncio.to_thread(self._p.kill_sandbox, sandbox_id)

    async def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[AsyncSandbox, SandboxInfo]:
        sb, info = await asyncio.to_thread(self._p.create_sandbox, req)
        assert isinstance(sb, LocalSandbox)
        return AsyncLocalSandbox(sb), info

    async def attach_sandbox(self, sandbox_id: str) -> AsyncSandbox:
        sb = await asyncio.to_thread(self._p.attach_sandbox, sandbox_id)
        assert isinstance(sb, LocalSandbox)
        return AsyncLocalSandbox(sb)

    async def close(self) -> None:
        return await asyncio.to_thread(self._p.close)


def _make_local_provider(cfg: ProviderConfig) -> Provider:
    return LocalProvider(cfg)


def _make_async_local_provider(cfg: ProviderConfig) -> AsyncProvider:
    return AsyncLocalProvider(cfg)


register_provider(ProviderName.LOCAL, _make_local_provider, _make_async_local_provider)



from . import e2b as _e2b
from . import daytona as _daytona
from . import blaxel as _blaxel
from . import runloop as _runloop
from . import fly_machines as _fly_machines
from . import local as _local



from __future__ import annotations

import base64
import os
import time
from datetime import datetime
from urllib.parse import quote, urlencode

from ..types import (
    ProviderName,
    SandboxStatus,
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)
from ..config import ProviderConfig
from ..provider import Provider, Sandbox
from ..errors import NotFoundError, NotSupportedError, ProviderError, BadConfigError
from .._http_client import HttpClient, HTTPError
from .._connect_proto import unary_post, stream_post, CONNECT_TIMEOUT_MS
from .._util import first_non_empty, shell_quote
from ..registry import register_provider

_DEFAULT_API_BASE = "https://api.e2b.app"
_DEFAULT_ENVD_PORT = 49983
_HEADER_API_KEY = "X-API-Key"
_HEADER_ACCESS_TOKEN = "X-Access-Token"


def _elapsed_ms(start: float) -> int:
    return int((time.monotonic() - start) * 1000)


def _map_err(name: str, err: HTTPError) -> Exception:
    if err.status == 404:
        return NotFoundError(str(err))
    msg = err.body.decode(errors="replace")
    if len(msg) > 256:
        msg = msg[:256] + "..."
    return ProviderError(provider=name, message=msg, status_code=err.status)


def _metadata_match(m: dict[str, str], needle: str) -> bool:
    if not needle:
        return True
    for v in m.values():
        if needle in v:
            return True
    return False


class E2BProvider(Provider):
    def __init__(self, cfg: ProviderConfig) -> None:
        key = first_non_empty(cfg.api_key, os.environ.get("E2B_API_KEY"))
        if not key:
            raise BadConfigError("E2B API key required (api_key or E2B_API_KEY)")
        timeout_s = 30.0
        if cfg.default_timeout_ms:
            timeout_s = cfg.default_timeout_ms / 1000.0
        self._hc = HttpClient(timeout_s=timeout_s)
        self._api_key = key
        self._api_base = (cfg.base_url or _DEFAULT_API_BASE).rstrip("/")
        port_str = os.environ.get("E2B_ENVD_PORT", "")
        self._port = int(port_str) if port_str.isdigit() else _DEFAULT_ENVD_PORT
        self._tpl = os.environ.get("E2B_TEMPLATE_ID", "base")

    def _api_headers(self) -> dict[str, str]:
        return {_HEADER_API_KEY: self._api_key}

    def _envd_base(self, sandbox_id: str) -> str:
        return f"https://{self._port}-{sandbox_id}.e2b.app"

    def _envd_headers(self, token: str | None) -> dict[str, str]:
        h: dict[str, str] = {}
        if token:
            h[_HEADER_ACCESS_TOKEN] = token
        return h

    def _get_sandbox_detail(self, sandbox_id: str) -> dict:
        url = f"{self._api_base}/sandboxes/{quote(sandbox_id, safe='')}"
        try:
            result = self._hc.do("GET", url, self._api_headers())
        except HTTPError as e:
            raise _map_err("e2b", e) from e
        return result  # type: ignore[return-value]

    def _detail_to_info(self, d: dict, fallback_tpl: str, token: str | None = None) -> SandboxInfo:
        tpl = d.get("templateID") or fallback_tpl
        state = d.get("state", "running")
        status = SandboxStatus.PAUSED if state == "paused" else SandboxStatus.RUNNING
        started = datetime.now()
        if d.get("startedAt"):
            try:
                started = datetime.fromisoformat(d["startedAt"].replace("Z", "+00:00"))
            except (ValueError, TypeError):
                pass
        expires: datetime | None = None
        if d.get("endAt"):
            try:
                expires = datetime.fromisoformat(d["endAt"].replace("Z", "+00:00"))
            except (ValueError, TypeError):
                pass
        cpus = d.get("cpuCount") or None
        memory = d.get("memoryMB") or None
        return SandboxInfo(
            id=d.get("sandboxID", ""),
            provider=ProviderName.E2B,
            template=tpl,
            status=status,
            started_at=started,
            expires_at=expires,
            metadata=d.get("metadata") or {},
            cpus=cpus if cpus and cpus > 0 else None,
            memory_mb=memory if memory and memory > 0 else None,
        )

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        filter = filter or ListSandboxesFilter()
        if filter.provider is not None and filter.provider != ProviderName.E2B:
            return []
        url = f"{self._api_base}/v2/sandboxes"
        if filter.limit and filter.limit > 0:
            url += f"?limit={filter.limit}"
        try:
            listed = self._hc.do("GET", url, self._api_headers())
        except HTTPError as e:
            raise _map_err("e2b", e) from e
        if not listed:
            return []
        out: list[SandboxInfo] = []
        for s in listed:  # type: ignore[union-attr]
            metadata = s.get("metadata") or {}
            if filter.metadata_filter and not _metadata_match(metadata, filter.metadata_filter):
                continue
            state = s.get("state", "running")
            status = SandboxStatus.PAUSED if state == "paused" else SandboxStatus.RUNNING
            started = datetime.now()
            if s.get("startedAt"):
                try:
                    started = datetime.fromisoformat(s["startedAt"].replace("Z", "+00:00"))
                except (ValueError, TypeError):
                    pass
            expires: datetime | None = None
            if s.get("endAt"):
                try:
                    expires = datetime.fromisoformat(s["endAt"].replace("Z", "+00:00"))
                except (ValueError, TypeError):
                    pass
            cpus = s.get("cpuCount") or None
            memory = s.get("memoryMB") or None
            out.append(
                SandboxInfo(
                    id=s.get("sandboxID", ""),
                    provider=ProviderName.E2B,
                    template=s.get("templateID"),
                    status=status,
                    started_at=started,
                    expires_at=expires,
                    metadata=metadata,
                    cpus=cpus if cpus and cpus > 0 else None,
                    memory_mb=memory if memory and memory > 0 else None,
                )
            )
        return out

    def kill_sandbox(self, sandbox_id: str) -> None:
        url = f"{self._api_base}/sandboxes/{quote(sandbox_id, safe='')}"
        try:
            self._hc.do("DELETE", url, self._api_headers())
        except HTTPError as e:
            raise _map_err("e2b", e) from e

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        req = req or CreateSandboxRequest()
        tpl = self._tpl
        if req.template:
            tpl = req.template
        body: dict = {
            "templateID": tpl,
            "metadata": req.metadata,
            "envVars": req.envs,
        }
        if req.timeout_seconds is not None:
            body["timeout"] = req.timeout_seconds
        if req.cpus is not None:
            body["cpuCount"] = req.cpus
        if req.memory_mb is not None:
            body["memoryMB"] = req.memory_mb
        url = f"{self._api_base}/sandboxes"
        try:
            created = self._hc.do("POST", url, self._api_headers(), body)
        except HTTPError as e:
            raise _map_err("e2b", e) from e
        sandbox_id = created["sandboxID"]  # type: ignore[index]
        token = created.get("envdAccessToken")  # type: ignore[union-attr]
        detail = self._get_sandbox_detail(sandbox_id)
        info = self._detail_to_info(detail, tpl, token)
        sb = E2BSandbox(provider=self, sandbox_id=sandbox_id, token=token)
        return sb, info

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        detail = self._get_sandbox_detail(sandbox_id)
        token = detail.get("envdAccessToken")
        return E2BSandbox(provider=self, sandbox_id=sandbox_id, token=token)

    def close(self) -> None:
        self._hc.close()


class E2BSandbox(Sandbox):
    def __init__(self, provider: E2BProvider, sandbox_id: str, token: str | None = None) -> None:
        self._provider = provider
        self._id = sandbox_id
        self._token = token

    @property
    def id(self) -> str:
        return self._id

    def info(self) -> SandboxInfo:
        detail = self._provider._get_sandbox_detail(self._id)
        return self._provider._detail_to_info(detail, self._provider._tpl, self._token)

    def is_running(self) -> bool:
        detail = self._provider._get_sandbox_detail(self._id)
        return detail.get("state") == "running"

    def pause(self) -> None:
        url = f"{self._provider._api_base}/sandboxes/{quote(self._id, safe='')}/pause"
        try:
            self._provider._hc.do("POST", url, self._provider._api_headers(), {})
        except HTTPError as e:
            raise _map_err("e2b", e) from e

    def resume(self) -> None:
        url = f"{self._provider._api_base}/sandboxes/{quote(self._id, safe='')}/resume"
        try:
            self._provider._hc.do("POST", url, self._provider._api_headers(), {})
        except HTTPError as e:
            raise _map_err("e2b", e) from e

    def kill(self) -> None:
        self._provider.kill_sandbox(self._id)

    def port_url(self, port: int) -> str:
        return f"https://{port}-{self._id}.e2b.app"

    def run_command(self, req: RunCommandRequest) -> CommandResult:
        start = time.monotonic()
        url = self._provider._envd_base(self._id) + "/process.Process/Start"
        h = self._provider._envd_headers(self._token)
        if req.timeout_seconds is not None and req.timeout_seconds > 0:
            h[CONNECT_TIMEOUT_MS] = str(req.timeout_seconds * 1000)
        proc: dict = {
            "cmd": "sh",
            "args": ["-c", req.cmd],
            "envs": req.env,
        }
        if req.cwd:
            proc["cwd"] = req.cwd
        body = {"process": proc}
        stdout_parts: list[str] = []
        stderr_parts: list[str] = []
        exit_code = -1
        saw_end = False

        def on_msg(msg: dict) -> None:
            nonlocal exit_code, saw_end
            event = msg.get("event")
            if not isinstance(event, dict):
                return
            data = event.get("data")
            if isinstance(data, dict):
                _append_b64(stdout_parts, data.get("stdout"))
                _append_b64(stderr_parts, data.get("stderr"))
            end = event.get("end")
            if isinstance(end, dict):
                ec = end.get("exitCode", end.get("exit_code", 0))
                exit_code = ec if isinstance(ec, int) else 0
                saw_end = True

        stream_post(self._provider._hc, url, h, body, on_msg)
        if not saw_end:
            exit_code = -1
        return CommandResult(
            stdout="".join(stdout_parts),
            stderr="".join(stderr_parts),
            exit_code=exit_code,
            duration_ms=_elapsed_ms(start),
        )

    def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        raise NotSupportedError("E2B does not support start_command")

    def wait_for_handle(self, handle_id: str) -> CommandResult:
        raise NotSupportedError("E2B does not support wait_for_handle")

    def kill_process(self, pid: int) -> None:
        raise NotSupportedError("E2B does not support kill_process")

    def list_processes(self) -> list[ProcessInfo]:
        raise NotSupportedError("E2B does not support list_processes")

    def read_file(self, path: str) -> bytes:
        url = self._provider._envd_base(self._id) + "/files?" + urlencode({"path": path})
        h = self._provider._envd_headers(self._token)
        try:
            _, data = self._provider._hc.do_raw("GET", url, h)
        except HTTPError as e:
            if e.status == 404:
                raise NotFoundError(f"file not found: {path}") from e
            raise _map_err("e2b", e) from e
        return data

    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        url = self._provider._envd_base(self._id) + "/files"
        h = self._provider._envd_headers(self._token)
        # multipart upload
        import io

        boundary = "----SandboxerBoundary"
        body = io.BytesIO()
        body.write(f"--{boundary}\r\n".encode())
        body.write(b'Content-Disposition: form-data; name="path"\r\n\r\n')
        body.write(path.encode())
        body.write(b"\r\n")
        body.write(f"--{boundary}\r\n".encode())
        body.write(b'Content-Disposition: form-data; name="file"; filename="blob"\r\n')
        body.write(b"Content-Type: application/octet-stream\r\n\r\n")
        body.write(content)
        body.write(b"\r\n")
        body.write(f"--{boundary}--\r\n".encode())
        h["Content-Type"] = f"multipart/form-data; boundary={boundary}"
        try:
            self._provider._hc.do_raw("POST", url, h, content=body.getvalue())
        except HTTPError as e:
            raise _map_err("e2b", e) from e

    def list_directory(self, path: str) -> list[FileInfo]:
        url = self._provider._envd_base(self._id) + "/filesystem.Filesystem/ListDir"
        h = self._provider._envd_headers(self._token)
        result = unary_post(self._provider._hc, url, h, {"path": path, "depth": 1})
        if not result:
            return []
        entries = result.get("entries", []) if isinstance(result, dict) else []
        out: list[FileInfo] = []
        for e in entries:
            is_dir = "DIRECTORY" in (e.get("type", "") or "")
            size = 0
            raw_size = e.get("size")
            if isinstance(raw_size, (int, float)):
                size = int(raw_size)
            elif isinstance(raw_size, str) and raw_size.isdigit():
                size = int(raw_size)
            file_mode = e.get("mode") or None
            out.append(
                FileInfo(
                    name=e.get("name", ""),
                    path=e.get("path", ""),
                    is_dir=is_dir,
                    size=size,
                    mode=file_mode if file_mode and file_mode != 0 else None,
                )
            )
        return out

    def make_dir(self, path: str) -> None:
        url = self._provider._envd_base(self._id) + "/filesystem.Filesystem/MakeDir"
        h = self._provider._envd_headers(self._token)
        unary_post(self._provider._hc, url, h, {"path": path})

    def remove(self, path: str) -> None:
        self.run_command(RunCommandRequest(cmd="rm -rf " + shell_quote(path)))

    def exists(self, path: str) -> bool:
        url = self._provider._envd_base(self._id) + "/files?" + urlencode({"path": path})
        h = self._provider._envd_headers(self._token)
        try:
            self._provider._hc.do_raw("GET", url, h)
        except HTTPError as e:
            if e.status == 404:
                return False
            raise _map_err("e2b", e) from e
        return True

    def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        raise NotSupportedError("E2B does not support create_pty")

    def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        raise NotSupportedError("E2B does not support resize_pty")

    def kill_pty(self, pid: int) -> None:
        raise NotSupportedError("E2B does not support kill_pty")

    def list_pty(self) -> list[PTYInfo]:
        raise NotSupportedError("E2B does not support list_pty")


def _append_b64(parts: list[str], val: str | None) -> None:
    if not val:
        return
    try:
        decoded = base64.b64decode(val).decode(errors="replace")
        parts.append(decoded)
    except Exception:
        parts.append(val)


def _make_e2b_provider(cfg: ProviderConfig) -> Provider:
    return E2BProvider(cfg)


register_provider(ProviderName.E2B, _make_e2b_provider)



from __future__ import annotations

import time
import asyncio
from dataclasses import dataclass
from typing import Callable, Awaitable


@dataclass
class RetryOptions:
    max_attempts: int = 1
    interval_s: float = 0.2


def do_retry(opts: RetryOptions, fn: Callable[[], None]) -> None:
    max_attempts = max(opts.max_attempts, 1)
    last_err: Exception | None = None
    for i in range(max_attempts):
        try:
            fn()
            return
        except Exception as e:
            last_err = e
        if i < max_attempts - 1:
            time.sleep(opts.interval_s)
    if last_err:
        raise last_err


async def async_do_retry(opts: RetryOptions, fn: Callable[[], Awaitable[None]]) -> None:
    max_attempts = max(opts.max_attempts, 1)
    last_err: Exception | None = None
    for i in range(max_attempts):
        try:
            await fn()
            return
        except Exception as e:
            last_err = e
        if i < max_attempts - 1:
            await asyncio.sleep(opts.interval_s)
    if last_err:
        raise last_err



from __future__ import annotations

from typing import Callable

from .types import ProviderName
from .config import ProviderConfig
from .provider import Provider, AsyncProvider
from .errors import BadConfigError

_sync_factories: dict[ProviderName, Callable[[ProviderConfig], Provider]] = {}
_async_factories: dict[ProviderName, Callable[[ProviderConfig], AsyncProvider]] = {}


def register_provider(
    name: ProviderName,
    sync_factory: Callable[[ProviderConfig], Provider],
    async_factory: Callable[[ProviderConfig], AsyncProvider] | None = None,
) -> None:
    _sync_factories[name] = sync_factory
    if async_factory:
        _async_factories[name] = async_factory


def resolve_provider(name: ProviderName, config: ProviderConfig | None = None) -> Provider:
    factory = _sync_factories.get(name)
    if not factory:
        raise BadConfigError(f'unknown provider "{name}"')
    return factory(config or ProviderConfig())


def resolve_async_provider(name: ProviderName, config: ProviderConfig | None = None) -> AsyncProvider:
    factory = _async_factories.get(name)
    if not factory:
        raise BadConfigError(f'unknown async provider "{name}"')
    return factory(config or ProviderConfig())



from __future__ import annotations

import os
import time
from datetime import datetime
from urllib.parse import quote

from ..types import (
    ProviderName,
    SandboxStatus,
    SandboxInfo,
    CommandResult,
    FileInfo,
    ProcessInfo,
    PTYInfo,
    CreateSandboxRequest,
    RunCommandRequest,
    StartCommandRequest,
    CreatePTYRequest,
    ListSandboxesFilter,
)
from ..config import ProviderConfig
from ..provider import Provider, Sandbox
from ..errors import NotFoundError, NotSupportedError, ProviderError, BadConfigError
from .._http_client import HttpClient, HTTPError
from .._util import first_non_empty
from ..registry import register_provider

_DEFAULT_API = "https://api.machines.dev"


def _map_err(err: HTTPError) -> Exception:
    if err.status == 404:
        return NotFoundError(str(err))
    msg = err.body.decode(errors="replace")
    if len(msg) > 400:
        msg = msg[:400] + "..."
    return ProviderError(provider="fly-machines", message=msg, status_code=err.status)


class FlyMachinesProvider(Provider):
    def __init__(self, cfg: ProviderConfig) -> None:
        tok = first_non_empty(cfg.api_key, os.environ.get("FLY_API_TOKEN"))
        if not tok:
            raise BadConfigError("Fly API token required (api_key or FLY_API_TOKEN)")
        app = os.environ.get("FLY_APP_NAME") or os.environ.get("SANDBOXER_FLY_APP") or ""
        if not app:
            raise BadConfigError("set FLY_APP_NAME or SANDBOXER_FLY_APP")
        timeout_s = 30.0
        if cfg.default_timeout_ms:
            timeout_s = cfg.default_timeout_ms / 1000.0
        self._hc = HttpClient(timeout_s=timeout_s)
        self._token = tok
        base = cfg.base_url or os.environ.get("FLY_API_HOSTNAME") or _DEFAULT_API
        self._base = base.rstrip("/")
        self._app = app

    def _hdr(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self._token}"}

    def list_sandboxes(self, filter: ListSandboxesFilter | None = None) -> list[SandboxInfo]:
        filter = filter or ListSandboxesFilter()
        if filter.provider is not None and filter.provider != ProviderName.FLY_MACHINES:
            return []
        url = f"{self._base}/v1/apps/{quote(self._app, safe='')}/machines"
        try:
            result = self._hc.do("GET", url, self._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        if not result:
            return []
        # API may return list directly or wrapped
        machines = result if isinstance(result, list) else result.get("machines", [])
        out: list[SandboxInfo] = []
        for m in machines:
            state = (m.get("state") or "running").lower()
            st = SandboxStatus.STOPPED if state in ("stopped", "destroyed") else SandboxStatus.RUNNING
            out.append(
                SandboxInfo(
                    id=m.get("id", ""),
                    provider=ProviderName.FLY_MACHINES,
                    status=st,
                    started_at=datetime.now(),
                    metadata={"region": m.get("region", ""), "app": self._app},
                )
            )
            if filter.limit and filter.limit > 0 and len(out) >= filter.limit:
                break
        return out

    def kill_sandbox(self, sandbox_id: str) -> None:
        url = f"{self._base}/v1/apps/{quote(self._app, safe='')}/machines/{quote(sandbox_id, safe='')}?force=true"
        try:
            self._hc.do("DELETE", url, self._hdr())
        except HTTPError as e:
            raise _map_err(e) from e

    def create_sandbox(self, req: CreateSandboxRequest | None = None) -> tuple[Sandbox, SandboxInfo]:
        req = req or CreateSandboxRequest()
        image = req.template or "nginx:alpine"
        cpus = req.cpus or 1
        mem = req.memory_mb or 256
        body = {
            "config": {
                "image": image,
                "guest": {
                    "cpu_kind": "shared",
                    "cpus": cpus,
                    "memory_mb": mem,
                },
                "auto_destroy": True,
                "auto_start_machines": True,
                "restart": {"policy": "no"},
                "stop_timeout": "60s",
                "env": req.envs,
                "metadata": req.metadata,
            },
            "region": first_non_empty(os.environ.get("FLY_REGION"), "iad"),
        }
        url = f"{self._base}/v1/apps/{quote(self._app, safe='')}/machines"
        try:
            created = self._hc.do("POST", url, self._hdr(), body)
        except HTTPError as e:
            raise _map_err(e) from e
        cid = created.get("id", "")  # type: ignore[union-attr]
        sb = FlyMachinesSandbox(provider=self, sandbox_id=cid)
        info = SandboxInfo(
            id=cid,
            provider=ProviderName.FLY_MACHINES,
            status=SandboxStatus.STARTING,
            started_at=datetime.now(),
            template=image,
            metadata={"app": self._app},
            cpus=cpus,
            memory_mb=mem,
        )
        return sb, info

    def attach_sandbox(self, sandbox_id: str) -> Sandbox:
        sb = FlyMachinesSandbox(provider=self, sandbox_id=sandbox_id)
        sb.info()  # validate
        return sb

    def close(self) -> None:
        self._hc.close()


class FlyMachinesSandbox(Sandbox):
    def __init__(self, provider: FlyMachinesProvider, sandbox_id: str) -> None:
        self._p = provider
        self._id = sandbox_id

    @property
    def id(self) -> str:
        return self._id

    def info(self) -> SandboxInfo:
        url = f"{self._p._base}/v1/apps/{quote(self._p._app, safe='')}/machines/{quote(self._id, safe='')}"
        try:
            m = self._p._hc.do("GET", url, self._p._hdr())
        except HTTPError as e:
            raise _map_err(e) from e
        state = (m.get("state") or "running").lower()  # type: ignore[union-attr]
        st = SandboxStatus.STOPPED if state == "stopped" else SandboxStatus.RUNNING
        config = m.get("config", {}) if isinstance(m, dict) else {}  # type: ignore[union-attr]
        guest = config.get("guest", {})
        image = config.get("image") or None
        cpus = guest.get("cpus") or None
        memory = guest.get("memory_mb") or None
        return SandboxInfo(
            id=m.get("id", ""),  # type: ignore[union-attr]
            provider=ProviderName.FLY_MACHINES,
            status=st,
            started_at=datetime.now(),
            template=image,
            metadata={"app": self._p._app},
            cpus=cpus if cpus and cpus > 0 else None,
            memory_mb=memory if memory and memory > 0 else None,
        )

    def is_running(self) -> bool:
        return self.info().status == SandboxStatus.RUNNING

    def pause(self) -> None:
        url = f"{self._p._base}/v1/apps/{quote(self._p._app, safe='')}/machines/{quote(self._id, safe='')}/suspend"
        try:
            self._p._hc.do("POST", url, self._p._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def resume(self) -> None:
        url = f"{self._p._base}/v1/apps/{quote(self._p._app, safe='')}/machines/{quote(self._id, safe='')}/start"
        try:
            self._p._hc.do("POST", url, self._p._hdr(), {})
        except HTTPError as e:
            raise _map_err(e) from e

    def kill(self) -> None:
        self._p.kill_sandbox(self._id)

    def port_url(self, port: int) -> str:
        raise NotSupportedError("Fly Machines does not support port_url")

    def run_command(self, req: RunCommandRequest) -> CommandResult:
        raise NotSupportedError("Fly Machines does not support run_command")

    def start_command(self, req: StartCommandRequest) -> tuple[int, str]:
        raise NotSupportedError("Fly Machines does not support start_command")

    def wait_for_handle(self, handle_id: str) -> CommandResult:
        raise NotSupportedError("Fly Machines does not support wait_for_handle")

    def kill_process(self, pid: int) -> None:
        raise NotSupportedError("Fly Machines does not support kill_process")

    def list_processes(self) -> list[ProcessInfo]:
        raise NotSupportedError("Fly Machines does not support list_processes")

    def read_file(self, path: str) -> bytes:
        raise NotSupportedError("Fly Machines does not support read_file")

    def write_file(self, path: str, content: bytes, mode: int | None = None, user: str | None = None) -> None:
        raise NotSupportedError("Fly Machines does not support write_file")

    def list_directory(self, path: str) -> list[FileInfo]:
        raise NotSupportedError("Fly Machines does not support list_directory")

    def make_dir(self, path: str) -> None:
        raise NotSupportedError("Fly Machines does not support make_dir")

    def remove(self, path: str) -> None:
        raise NotSupportedError("Fly Machines does not support remove")

    def exists(self, path: str) -> bool:
        raise NotSupportedError("Fly Machines does not support exists")

    def create_pty(self, req: CreatePTYRequest) -> PTYInfo:
        raise NotSupportedError("Fly Machines does not support create_pty")

    def resize_pty(self, pid: int, rows: int, cols: int) -> None:
        raise NotSupportedError("Fly Machines does not support resize_pty")

    def kill_pty(self, pid: int) -> None:
        raise NotSupportedError("Fly Machines does not support kill_pty")

    def list_pty(self) -> list[PTYInfo]:
        raise NotSupportedError("Fly Machines does not support list_pty")


def _make_fly_provider(cfg: ProviderConfig) -> Provider:
    return FlyMachinesProvider(cfg)


register_provider(ProviderName.FLY_MACHINES, _make_fly_provider)
